//********************************************************************
//*** ResourcePool/mysql/SQLSelect.hh
//*** Copyright (c) 2003-2009 by Markus Winand <mws@fatalmind.com>
//*** $Id: SQLSelect.cc,v 1.32 2009-03-07 23:04:02 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */


#include "SQLSelect.hh"

#include "lex.mysql_bind.hh"

#ifndef INCLUDED_EXCEPTIONS_HH
#include "Exceptions/Exception.hh"
#endif

#ifndef RESOURCEPOOL_MYSQL_SQLTYPE_HH
#include "SQLType.hh"
#endif

#ifndef RESOURCEPOOL_MYSQL_MYSQLEXCEPTION_HH
#include "MysqlException.hh"
#endif

#include "ResourcePool/SQL/BindOut.hh"

#include <mysql/mysql.h>
#include <sstream>

namespace fatalmind {
namespace mysql {

SQLSelect::SQLSelect(const std::string& SQL, SQL::SQLFetcher& fetcher)
: super(fetcher)
, _SQL(SQL)
, highlimit(0)
, lowlimit(0)
, hardhighlimit(false)
, _rs(0)
, _num_fields(0)
{
}

SQLSelect::SQLSelect(const ResourcePool<ResourceType>& p, const std::string& SQL, SQL::SQLFetcher& fetcher)
: super(fetcher)
, _SQL(SQL)
, highlimit(0)
, lowlimit(0)
, hardhighlimit(false)
, _rs(0)
, _num_fields(0)
{
}

SQLSelect::~SQLSelect()
{
}

void SQLSelect::finish() {
	if (_rs) {
		while (mysql_fetch_row(_rs)) {
			// empty buffer (whow, thats really required)
		}
		close_rs();
	}
}

void SQLSelect::close_rs() {
	mysql_free_result(_rs);
	_rs = 0;
}

void SQLSelect::execute(mysqlResource& rdbh)
{
	MYSQL& dbh = rdbh.get_plain_resource();
	finish();
	mysql_bind boundsql(dbh, _SQL, _args);
	if (int rc = mysql_real_query(&dbh
		, boundsql.getSQL().c_str()
		, boundsql.getSQL().length())
	) {
		(*rdbh.getReader(rdbh.Event.EXECUTE)) << mysql_errno(&dbh) << ";" << (*this) ;
		MysqlException::throwit(rc, mysql_error(&dbh), mysql_errno(&dbh), _SQL);
	}
	
	if ((_rs = mysql_use_result(&dbh)) == NULL) {
		(*rdbh.getReader(rdbh.Event.EXECUTE)) << mysql_errno(&dbh) << ";" << (*this) ;
		MysqlException::throwit(0, mysql_error(&dbh), mysql_errno(&dbh), _SQL);
	}
	(*rdbh.getReader(rdbh.Event.EXECUTE)) << ';' << (*this);
	_num_fields = mysql_num_fields(_rs);

	internal::registerFinish fin(*this);
	
	bool more = true;
	int  row  = 0;
	while ((more = fetch()) && ishighlimit(row, rdbh) && ++row && fetcher.process(row)) {
		rdbh.getReader(rdbh.Event.FETCH);
	}
	islowlimit(row, rdbh);
	fetcher.done(more, row);
}

bool SQLSelect::islowlimit(int rows, mysqlResource& rdbh) const {
	if (lowlimit > 0 && rows < lowlimit) {
		std::ostringstream os;
		os << "The resultset has less records then expected (expected not less then " << lowlimit << ")";
		(*rdbh.getReader(rdbh.Event.FETCH)) << "NO DATA;" << (*this);
		throw NFOMysqlException(os.str(), 99, _SQL);
	}
}

bool SQLSelect::ishighlimit(int rows, mysqlResource& rdbh) const {
	if ((highlimit > 0) && (rows >= highlimit)) {
		if (hardhighlimit) {
			std::ostringstream os;
			os << "The resultset has more records then expected (expected not more then " << highlimit << ")";
			(*rdbh.getReader(rdbh.Event.FETCH)) << "MORE DATA;" << (*this);
			throw NFOMysqlException(os.str(), 1172, _SQL);
		} else {
			return false;
		}
	}
	return true;
}

bool SQLSelect::fetch() {
	if (_rs) {
		// TODO: improve exception safty by introducing an object for _rs
		try {
			MYSQL_ROW row;
			row = mysql_fetch_row(_rs);
			if (row) {
				unsigned long *lengths;
				lengths = mysql_fetch_lengths(_rs);
				if (_outparam.size() < _num_fields) {
					std::ostringstream os;
					os << "bindout not called for all " << _num_fields << " columns.";
					throw NFOMysqlException(os.str(), 98, _SQL);
				}
				for (unsigned int i = 0; i < _num_fields; ++i) {
					if (_outparam[i].get() != 0) {
						SQLRefType& op = *_outparam[i];
						op.assign(row[i], lengths[i]);
					} else {
						std::ostringstream os;
						os << "bindout not called for column " << i;
						throw NFOMysqlException(os.str(), 98, _SQL);
					}
				}
			} else {
				close_rs();
				return false;
			}
		} catch (...) {
			close_rs();
			throw;
		}
	} else {
		return false;
	}
	return true;
}

void SQLSelect::outputoperator(std::ostream& s) const {
	s << "mysql::SQLSelect[" << _SQL << '|';
	bindoutputoperator(s);
	s << ']';
}

} // namespace mysql

// 
// char types
//
template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<char>
	(const unsigned int pos, char& val, bool& isnull);
template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<char>
	(const unsigned int pos, char& val, BindOutNotNullConstraint);

template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<signed char>
	(const unsigned int pos, signed char& val, bool& isnull);
template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<signed char>
	(const unsigned int pos, signed char& val, BindOutNotNullConstraint);

template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<unsigned char>
	(const unsigned int pos, unsigned char& val, bool& isnull);
template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<unsigned char>
	(const unsigned int pos, unsigned char& val,  BindOutNotNullConstraint);

//
// int types
//

template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<short int>
	(const unsigned int pos, short int& val, bool& isnull);
template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<short int>
	(const unsigned int pos, short int& val, BindOutNotNullConstraint);

template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<int>
	(const unsigned int pos, int& val, bool& isnull);
template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<int>
	(const unsigned int pos, int& val, BindOutNotNullConstraint);

template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<long int>
	(const unsigned int pos, long int& val, bool& isnull);
template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<long int>
	(const unsigned int pos, long int& val, BindOutNotNullConstraint);

template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<unsigned short int>
	(const unsigned int pos, unsigned short int& val, bool& isnull);
template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<unsigned short int>
	(const unsigned int pos, unsigned short int& val, BindOutNotNullConstraint);
template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<unsigned int>
	(const unsigned int pos, unsigned int& val, bool& isnull);
template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<unsigned int>
	(const unsigned int pos, unsigned int& val, BindOutNotNullConstraint);
template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<unsigned long int>
	(const unsigned int pos, unsigned long int& val, bool& isnull);
template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<unsigned long int>
	(const unsigned int pos, unsigned long int& val, BindOutNotNullConstraint);

//
// float types
//

template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<float>
	(const unsigned int pos, float& val, bool& isnull);
template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<float>
	(const unsigned int pos, float& val, BindOutNotNullConstraint);

template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<double>
	(const unsigned int pos, double& val, bool& isnull);
template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<double>
	(const unsigned int pos, double& val, BindOutNotNullConstraint);

template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<long double>
	(const unsigned int pos, long double& val, bool& isnull);
template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<long double>
	(const unsigned int pos, long double& val, BindOutNotNullConstraint);

//
// std::string
//
template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<std::string>
	(const unsigned int pos, std::string& val, bool& isnull);

template void SQL::BindOut<mysql::SQLRefType, mysql::SQLRefTypeImpl>::bindout<std::string>
	(const unsigned int pos, std::string& val, BindOutNotNullConstraint);

} // namespace fatalmind
